# coding: utf-8

"""
.. .. rst-class:: html-toggle

Abstract
========

A module to send ``raccomandate``, ``telegrammi`` and ``letters``
throught Italian Mail Service poste-online_.

A simple and tipical example could be as simple as::

    # Create a client for each type of letter
    xol = poste.ROLClient(username=USERNAME, password=PASSWORD)

    # Create a recipient and a Sender
    recipient = poste.Destinatario(**RECIPIENT)
    mittente = poste.Mittente(**MITTENTE)

    # send a document and confirm
    xol_submit = xol.submit('mydocument.pdf', recipient, mittente)
    result = xol.confirm(xol_submit=xol_submit, autoconfirm=True)

Three different clients are available:

* :class:`poste.ROLClient`: client to send ``raccomandate``

* :class:`poste.LOLClient`: client to send letters

* :class:`poste.TOLClient`: client to send telegrams

Test :class:`poste.xol.ROLSubmit`

.. .. rst-class:: html-toggle

API
====

.. rst-class:: html-toggle

Lol
------

.. autoclass:: poste.LOLClient
   :members:

Rol
----

.. autoclass:: poste.ROLClient
   :members:
   :show-inheritance:

Tol
----

.. autoclass:: poste.TOLClient
   :members:

.. rst-class:: html-toggle

Utility
--------

.. automodule:: poste.xol
   :members:

.. _poste-online: http://www.posta-online.it

.. rst-class:: html-toggle

Eccezioni
----------

.. autoexception:: poste.PosteSoapClientError

.. autoexception:: poste.ValidationError

.. autoexception:: poste.RemoteValidationError

.. autoexception:: poste.RemoteRecipientValidationError

.. autoexception:: poste.RemoteValidationError

.. autoexception:: poste.RecipientValidationError


"""
from __future__ import unicode_literals

import logging

from poste.tol import *
from poste.xol import *
from poste.webservice import *
from .utils import *


class PosteSoapClientError(Exception):
    """
    Un errore sollevato dal servizio remoto di poste-online
    """

    def __init__(self, result):
        super(PosteSoapClientError, self).__init__()
        try:
            self.servizio_enquiry = result.ServizioEnquiryResponse
        except:
            pass

        self.ceresult = result.CEResult
        self.cer_code = self.ceresult.Code
        self.cer_type = self.ceresult.Type
        self.cer_description = self.ceresult.Description

    def __str__(self):
        return self.cer_description


class ValidationError(Exception):
    """
    Errore generico di validazione
    """

    def __init__(self, msg, obj=None):
        # repr è necessario nel caso il testo abbia delle lettere accentate
        super(ValidationError, self).__init__(repr(msg))
        self.obj = obj


class RemoteValidationError(ValidationError):
    """
    Errore di validazione sollevato dal servizio poste-online
    """

    def __init__(self, result, guid):
        """
        :param result: remote result
        :param guid: guid
        """

        super(Exception, self).__init__()
        self.guid = guid
        try:
            self.servizio_enquiry = result.ServizioEnquiryResponse
        except:
            pass

        if hasattr(result, 'CEResult'):
            self.ceresult = result.CEResult
            self.cer_code = self.ceresult.Code
            self.cer_type = self.ceresult.Type
        else:
            self.ceresult = result.Result
            self.cer_code = self.ceresult.Code
            self.cer_type = self.ceresult.ResType

        try:
            self.cer_description = result.ServizioEnquiryResponse[0].StatoLavorazione.Descrizione
        except:
            self.cer_description = self.ceresult.Description

    def __str__(self):
        return self.cer_description


class RecipientValidationError(ValidationError):
    """Errore di Validatione del ricevente

    :arg errors: list of errors. Can be inspected in ``self.errors``
    """

    def __init__(self, msg, errors=None):
        super(RecipientValidationError, self).__init__(msg=msg)
        self.errors = errors or []


class RemoteRecipientValidationError(RemoteValidationError):
    """
    Errore di validazione del ricevente sollevato dal servizio poste-online
    """
    pass


def debug(active=False, filename=None):
    """activate deactivate debug. Very basic aimed at silencing suds
    Warn: You may be better off configuring it by yourself

    :param active: boolean
    :param filename: filename where to log suds

    """

    if filename:
        logging.basicConfig(filename=filename, level=logging.DEBUG)
    if active:
        logging.basicConfig(level=logging.DEBUG)
        logging.getLogger('suds.client').setLevel(logging.DEBUG)
        logging.getLogger('suds.transport').setLevel(
            logging.DEBUG)  # MUST BE THIS?
        logging.getLogger('suds.xsd.schema').setLevel(logging.DEBUG)
        logging.getLogger('suds.wsdl').setLevel(logging.DEBUG)
        logging.getLogger('suds.resolver').setLevel(logging.DEBUG)
        logging.getLogger('suds.xsd.query').setLevel(logging.DEBUG)
        logging.getLogger('suds.xsd.basic').setLevel(logging.DEBUG)
        logging.getLogger('suds.binding.marshaller').setLevel(logging.DEBUG)
    else:
        logging.basicConfig(level=logging.ERROR)
        logging.getLogger('suds').setLevel(logging.ERROR)
        logging.getLogger('suds.client').setLevel(logging.ERROR)
        logging.getLogger('suds.transport').setLevel(
            logging.ERROR)  # MUST BE THIS?
        logging.getLogger('suds.xsd.schema').setLevel(logging.ERROR)
        logging.getLogger('suds.wsdl').setLevel(logging.ERROR)
        logging.getLogger('suds.resolver').setLevel(logging.ERROR)
        logging.getLogger('suds.xsd.query').setLevel(logging.ERROR)
        logging.getLogger('suds.xsd.basic').setLevel(logging.ERROR)
        logging.getLogger('suds.binding.marshaller').setLevel(logging.ERROR)
