# coding: utf-8
from __future__ import unicode_literals

import re

import poste


class RecipientTOL(object):
    """
    Destinatario e validazione
    """

    def __init__(self, cognome='', nome='', ragione_sociale='',
                 indirizzo='', citta='', cap='', forza_destinazione='',
                 stato='', telefono='', validate=True):
        """
        :param cognome: cognome
        :param nome: nome
        :param ragione_sociale:
        :param indirizzo: max 41 caratteri
        :param citta: max 36 caratteri
        :param cap: codice avviamento postale - max 5 caratteri
        :param forza_destinazione:
        :param stato: richiesto solo per estero
        :param telefono:
        :param validate: boolean - if False validation will not be triggered automatically
        """
        self.cognome = cognome
        self.nome = nome
        self.ragione_sociale = ragione_sociale
        self.citta = citta
        self.cap = cap
        self.indirizzo = indirizzo
        self.stato = stato
        self.telefono = telefono
        self.forza_destinazione = forza_destinazione
        if validate:
            self.validate()

    def validate(self):
        """Check that all parameters fit what Poste requires"""
        # implementa pag. 18 di "Specifiche interfaccia"

        dest_name = " ".join([self.nome, self.cognome, self.ragione_sociale])
        if len(dest_name) > 41:
            msg = "Nome + cognome + Ragione sociale supera 41 caratteri (%s)" % dest_name
            raise poste.RecipientValidationError(msg)

        if not re.search('^[0-9]*$', self.telefono):
            msg = "Il telefono può contenere solo caratteri numerici (%s)" % dest_name
            raise poste.RecipientValidationError(msg)

        if len(self.telefono) > 12:
            msg = "Il telefono può contenere al più 12 caratteri (%s)" % dest_name
            raise poste.RecipientValidationError(msg)

        # FIXME: il telefono è necessario se TipoRecapito è valorizzato (paragr 3.8.3.5)
        #   ma tipo_recapito non è un attributo del Destinatario!!!

        if len(self.cap) > 5 or not self.cap:
            msg = "Il CAP può contenere al più 5 caratteri (%s - %s)" % (
                dest_name, self.cap)
            raise poste.RecipientValidationError(msg)

        if len(self.citta) > 36:
            msg = "La citta può contenere al più 36 caratteri (%s)" % dest_name
            raise poste.RecipientValidationError(msg)

        # 3.8.9.3 pag. 18
        if not self.cognome and (self.nome and not self.ragione_sociale):
            msg = "Cognome deve essere presente se manca nome ed è valorizzata " + \
                  "Ragione sociale (%r)" % (self)
            raise poste.RecipientValidationError(msg)

        if not self.cognome and (self.nome and not self.ragione_sociale):
            msg = "Il cognome è necessario se manca ragione sociale %s"
            raise poste.RecipientValidationError(msg % self)

        if not self.nome and (self.cognome and not self.ragione_sociale):
            msg = "Il nome è necessario se manca ragione sociale %s"
            raise poste.RecipientValidationError(msg % self)

    def get_instance(self, client):
        """Return an instance from WebService"""
        obj = client.type_factory("ns0").Recipient()
        obj.destinatario = client.type_factory("ns2").Destinatario()
        obj.destinatario.Telefono = self.telefono
        obj.destinatario.RagioneSociale = self.ragione_sociale
        obj.destinatario.Cognome = self.cognome
        obj.destinatario.Nome = self.nome
        obj.destinatario.Citta = self.citta
        obj.destinatario.CAP = self.cap
        obj.destinatario.Stato = self.stato
        obj.destinatario.ForzaDestinazione = self.forza_destinazione
        return obj

    def __repr__(self):
        return "<Dest: %s %s %s>" % (self.nome, self.cognome, self.ragione_sociale)


Destinatario = RecipientTOL
